;--------------------------------------------------------------------
; Module MEMSWAP.ASM
;
; Routines to swap memory to disk and spawn a subprocess.
; Follow memory chain past PSP, writing program blocks to disk
; and marking the blocks free.  Spawn program once memory is
; swapped out, read memory back in after spawn returns.
;
; Copyright (C) 1994, Brian Sawert.
; All rights reserved.
;
; External entry point:
;	int swap_spawn(char *cmd);
;
; History:
;	03/02/94 - BAS - original code
;
; Notes:
;	Code and data reside in same segment.
;	This module should appear near the beginning in the linker
;		command line, just after the startup module.
;	May require further safeguards - save DTA, vector table.
;	Command line must contain full path and file name of program
;	to spawn.
;--------------------------------------------------------------------

	DOSSEG
	LOCALS
	.MODEL SMALL


;-------------------- defines and equates --------------------

$DOS_INT		EQU 021h	; DOS interrupt
$DOS_GET_PSP		EQU 06200h	; DOS get PSP function
$DOS_GET_DRIVE		EQU 01900h	; DOS get drive function
$DOS_SET_DRIVE		EQU 00e00h	; DOS set drive function
$DOS_GET_DIR		EQU 04700h	; DOS get directory function
$DOS_SET_DIR		EQU 03b00h	; DOS set directory function
$DOS_FIND_FIRST		EQU 04e00h	; DOS directory search
$DOS_FILE_CREATE	EQU 03c00h	; DOS create file
$DOS_FILE_OPEN		EQU 03d00h	; DOS open file
$DOS_FILE_CLOSE		EQU 03e00h	; DOS close file
$DOS_FILE_DELETE	EQU 04100h	; DOS delete file
$DOS_FILE_WRITE		EQU 04000h	; DOS write to file
$DOS_FILE_READ		EQU 03f00h	; DOS read from file
$DOS_SET_ATTRIB		EQU 04301h	; DOS set file attribute
$DOS_PRINT_STRING	EQU 00900h	; DOS print string
$DOS_EXEC		EQU 04b00h	; DOS execute function
$DOS_EXEC_RETURN	EQU 04d00h	; DOS get execute return code
$DOS_EXIT		EQU 04c00h	; DOS exit program


$ENV_SEG_OFFSET		EQU 02ch	; environment segment offset in PSP
$PARA_SIZE		EQU 010h	; paragraph size
$FCB_SIZE		EQU 010h	; FCB size

$MAX_SWAP_ENTRIES	EQU 20		; max entries in swap table
$MAX_ENV_SIZE		EQU (128 * $PARA_SIZE)	; max environment size 
$MAX_PATH		EQU 128		; max path length
$COMMAND_LENGTH		EQU 80		; command line argument length
$READ_WRITE_PARAS	EQU 0400h	; read/write size


MCB		STRUC			; memory control block
	mcb_id		DB 0		; MCB ID ('M' or 'Z')
	owner		DW 0		; owner process ID (PSP of owner)
	nparas		DW 0		; length in paragraphs
			DB 3 DUP (0)	; reserved bytes
	progname	DB 8 DUP (0)	; owner name
MCB		ENDS

MEMSAVE		STRUC			; memory save structure
	memseg		DW ?		; memory segment address
	saveparas	DW ?		; paragraphs to save
MEMSAVE		ENDS

PARAMBLK	STRUC			; parameter block for child process
	env_addr	DW ?		; environment segment
	cmd_addr	DD ?		; address of command line tail
	fcb_addr_1	DD ?		; address of first default FCB
	fcb_addr_2	DD ?		; address of second default FCB
PARAMBLK	ENDS


;==================== start of code and data ====================

	.CODE
	ASSUME cs:@CurSeg, ds:@CurSeg, ss:@CurSeg


IFDEF	DEBUG				; data symbols
	PUBLIC	sentinel
	PUBLIC	cmd_length, cmd_line, cmd_tail
	PUBLIC	psp_seg, env_seg
	PUBLIC	new_env, new_env_seg, env_paras, tail_pad
	PUBLIC	swap_table, swap_count, save_count
	PUBLIC	cur_drive, cur_dir, swap_file
	PUBLIC	cmd_tail, blank_fcb, exec_param
	PUBLIC	spawn_code
ENDIF

IFDEF	DEBUG				; code symbols
	PUBLIC	relocate_env, restore_env
	PUBLIC	split_psp, restore_psp, build_table
	PUBLIC	save_dir, restore_dir
	PUBLIC	lock_swap, delete_swap
	PUBLIC	create_swap, open_swap, close_swap
	PUBLIC	write_mem, read_mem, spawn_proc
ENDIF


;==================== start of data ====================


EVENDATA
IFDEF	DEBUG
sentinel	DW 0abcdh		; stack sentinel
ENDIF
local_stack	DB 080h DUP (0)		; local stack setup
end_stack 	DW $			; end of local stack

dos_ss		DW ?			; saved stack values
dos_sp		DW ?
dos_bp		DW ?

psp_seg		DW ?			; local copy of PSP
env_seg		DW ?			; environment segment
new_env_seg	DW ?			; relocated environment segment
env_paras	DW ?			; env size in paragraphs

swap_count	DW 0			; swapped MCB count
save_count	DW 0			; saved MCB count
swap_table	MEMSAVE $MAX_SWAP_ENTRIES DUP (<>)
					; table of swapped block info
cur_drive	DB 0			; current disk drive number
cur_dir		DB '\', $MAX_PATH DUP (0)	; current working directory

swap_file	DB '_MEMSWAP.SWP', 0	; swap file name

spawn_code	DW ?			; exit code from system call

EVENDATA
cmd_length	DW ?			; actual length of command line
cmd_line	DB $COMMAND_LENGTH DUP (0)	; spawn command line
cmd_tail	DB $COMMAND_LENGTH DUP (0)	; spawn command tail
blank_fcb	DB ($FCB_SIZE * 2) DUP (0)	; default blank FCB space
exec_param	PARAMBLK <>		; parameter block for exec function

file_err_msg	DB 'Cannot open swap file - program aborted.', 0ah, 0dh, '$'



;==================== start of code ====================


;--------------------------------------------------------------------
; Relocate the environment block to the tail of the program stub.
;
; On entry:
;	BX = PSP segment
; 
; Returns:
;	Nothing
;--------------------------------------------------------------------

relocate_env		PROC
	push cx dx

	mov [env_paras], 0		; preset env size
	mov ax, OFFSET new_env + $PARA_SIZE - 1	; determine new env segment
	mov cx, 4
	shr ax, cl
	mov cx, ds
	add ax, cx
	mov [new_env_seg], ax		; save new segment
	mov dx, ax

	mov es, bx			; set ES to PSP segment
	mov ax, es:[$ENV_SEG_OFFSET]	; get environment segment
	mov [env_seg], ax		; save it
	cmp ax, bx			; is it before PSP segment?
	jb @@exit			; do not relocate

	mov es, ax
	sub di, di
	sub al, al
	mov cx, $MAX_ENV_SIZE

find_env_end:
	repne scasb			; search for first zero
	jne copy_env			; copy what we have

	cmp BYTE PTR es:[di], 0		; check for second zero
	jne find_env_end		; keep looking

copy_env:
	sub cx, $MAX_ENV_SIZE		; get environment size
	neg cx

	push ds
	push es
	pop ds
	sub si, si			; point to old environment
	mov es, dx			; retrieve new environment segment
	sub di, di			; point to new environment
	mov dx, cx			; save size
	rep movsb			; copy to new location
	sub ax, ax
	mov cx, 6
	add dx, cx
	rep stosb			; terminate environment
	mov ax, dx
	pop ds

	add ax, $PARA_SIZE - 1		; get environment size in paragraphs 
	mov cx, 4
	shr ax, cl
	mov [env_paras], ax		; save size

	mov dx, es
	mov es, bx
	mov es:[$ENV_SEG_OFFSET], dx	; set new environment segment

@@exit:
	pop dx cx
	ret
relocate_env		ENDP



;--------------------------------------------------------------------
; Restore the environment to its old location.  Set the segment
; value in the PSP to the old value.
;
; On entry:
;	BX = PSP segment
; 
; Returns:
;	Nothing
;--------------------------------------------------------------------

restore_env		PROC
	mov es, bx
	mov ax, [env_seg]		; get old environment segment
	mov es:[$ENV_SEG_OFFSET], ax 	; restore old value

	ret
restore_env		ENDP



;--------------------------------------------------------------------
; Split the PSP memory block, creating a swappable MCB at the end
; of the program stub.
;
; On entry:
;	BX = PSP segment
;	new_env_seg and env_paras values set
; 
; Returns:
;	Nothing
;--------------------------------------------------------------------

split_psp		PROC
	push cx dx si

	mov ax, [new_env_seg]		; determine tail MCB segment
	add ax, [env_paras]
	mov si, ax			; save tail MCB segment
	sub ax, bx			; calculate stub paragraphs

	mov dx, bx
	dec dx
	mov es, dx			; set ES to PSP MCB
	mov cx, es:[0].nparas		; get program size
	mov es:[0].nparas, ax		; set stub size
	mov dl, es:[0].mcb_id		; get MCB ID character
	mov es:[0].mcb_id, 'M'		; set MCB ID

	sub cx, ax			; calculate tail paragraphs
	dec cx
	mov es, si			; set ES to tail MCB
	mov es:[0].nparas, cx		; set tail size
	mov es:[0].owner, bx		; set tail owner
	mov es:[0].mcb_id, dl		; set tail ID

	pop si dx cx
	ret
split_psp		ENDP



;--------------------------------------------------------------------
; Restore the PSP memory block, rejoining the tail MCB to the
; original.
;
; On entry:
;	BX = PSP segment
;	tail_mcb value set
; 
; Returns:
;	Nothing
;--------------------------------------------------------------------

restore_psp		PROC
	push cx dx

	mov ax, [new_env_seg]		; determine tail MCB segment
	add ax, [env_paras]
	mov es, ax			; point to tail MCB

	mov cx, es:[0].nparas		; get tail paragraphs
	inc cx				; add MCB size
	mov dl, es:[0].mcb_id		; get tail ID

	mov ax, bx			; retrieve PSP segment
	dec ax
	mov es, ax			; point to PSP MCB
	mov ax, es:[0].nparas		; get PSP paragraph count
	add ax, cx
	mov es:[0].nparas, ax		; update paragraph count
	mov es:[0].mcb_id, dl		; update PSP ID

	pop dx cx
	ret
restore_psp		ENDP



;--------------------------------------------------------------------
; Follow memory allocation chain from PSP up and build swap table.
; Save the segment and length of program memory blocks.
;
; On entry:
;	BX = PSP segment
;	environment relocated, PSP block split
;
; Returns:
;	Nothing
;--------------------------------------------------------------------

build_table		PROC
	push cx dx si

	mov cx, $MAX_SWAP_ENTRIES	; set max entry count
	mov si, OFFSET swap_table

	mov dx, bx
	dec dx
	mov es, dx			; set ES to PSP MCB
	mov ax, es:[0].nparas		; get block size
	inc ax				; add MCB size

next_mcb:
	cmp es:[0].mcb_id, 'Z'		; check for end of chain
	je @@exit			; end of chain
	add dx, ax			; point to next MCB
	mov es, dx

	mov ax, es:[0].nparas		; get block size
	inc ax				; add MCB size
	cmp es:[0].owner, bx		; is it ours?
	jne next_mcb			; no - skip it

add_entry:
	mov [si].memseg, dx		; save memory segment
	mov [si].saveparas, ax		; save segment size
	add si, SIZE MEMSAVE		; point to next table entry
	dec cx
	jnz next_mcb			; room for more

@@exit:
	sub cx, $MAX_SWAP_ENTRIES
	neg cx
	mov [swap_count], cx		; save swap entry count
	
	pop si dx cx
	ret
build_table		ENDP



;--------------------------------------------------------------------
; Save current drive and directory.  Store in cur_drive and cur_dir.
;
; On entry:
;	Nothing
;
; Returns:
;	Nothing
;--------------------------------------------------------------------

save_dir		PROC
	push dx

	mov ax, $DOS_GET_DRIVE		; DOS get current drive function
	int $DOS_INT
	mov [cur_drive], al		; save current drive

	mov ax, $DOS_GET_DIR		; DOS get current directory function
	mov si, OFFSET cur_dir + 1	; space to save current directory
	sub dl, dl			; use default drive
	int $DOS_INT

	pop dx
	ret
save_dir		ENDP



;--------------------------------------------------------------------
; Restore current drive and directory.
;
; On entry:
;	cur_drive and cur_dir valid
;
; Returns:
;	Nothing
;--------------------------------------------------------------------

restore_dir		PROC
	push dx

	mov ax, $DOS_SET_DRIVE		; DOS set current drive function
	mov dl, [cur_drive]		
	int $DOS_INT

	mov ax, $DOS_SET_DIR		; DOS get current directory function
	mov dx, OFFSET cur_dir		; point to current directory
	int $DOS_INT

	pop dx
	ret
restore_dir		ENDP



;--------------------------------------------------------------------
; Set swap file read-only attribute.
;
; On entry:
;	AX = file attribute (1 to lock, 0 to unlock)
;	DX = pointer to file name
;
; Returns:
;	Nothing
;--------------------------------------------------------------------

lock_swap		PROC
	push cx

	mov cx, ax			; get file attribute
	mov ax, $DOS_SET_ATTRIB		; DOS set file attribute function
	int $DOS_INT

	pop cx
	ret
lock_swap		ENDP



;--------------------------------------------------------------------
; Delete swap file.
;
; On entry:
;	DX = pointer to file name
;
; Returns:
;	Nothing
;
; Note:
;	Uses default DTA in PSP for directory search.  If DTA has
;	moved, reset it to a buffer in program stub before swap.
;--------------------------------------------------------------------

delete_swap		PROC
	push cx

	sub cx, cx			; clear search attributes
	mov ax, $DOS_FIND_FIRST		; DOS directory search function
	int $DOS_INT
	jc @@exit			; no swap file found

	mov ax, 0
	call lock_swap			; unlock swap file

	mov ax, $DOS_FILE_DELETE	; DOS delete file function
	int $DOS_INT

@@exit:
	pop cx
	ret
delete_swap		ENDP



;--------------------------------------------------------------------
; Create swap file with read/write access.
;
; On entry:
;	DX = pointer to file name
;
; Returns:
;	Carry set on failure, clear on success with handle in BX
;--------------------------------------------------------------------

create_swap		PROC
	push cx 

	call delete_swap		; delete existing file

	mov ax, $DOS_FILE_CREATE	; DOS create file function
	sub cx, cx			; file attribute is read/write
	int $DOS_INT

	jnc @@exit
	mov ax, 0			; return 0 for handle

@@exit:
	mov bx, ax			; return file handle

	pop cx
	ret
create_swap		ENDP



;--------------------------------------------------------------------
; Open swap file for reading.
;
; On entry:
;	DX = pointer to file name
;
; Returns:
;	Carry set on failure, clear on success with handle in BX.
;--------------------------------------------------------------------

open_swap		PROC
	mov ax, $DOS_FILE_OPEN		; DOS open file function
	int $DOS_INT

	jnc @@exit
	mov ax, 0			; return 0 for handle

@@exit:
	mov bx, ax			; return file handle

	ret
open_swap		ENDP



;--------------------------------------------------------------------
; Close swap file.
;
; On entry:
;	BX = swap file handle
;
; Returns:
;	Nothing
;--------------------------------------------------------------------

close_swap		PROC
	mov ax, $DOS_FILE_CLOSE		; DOS close file function
	int $DOS_INT
	
	ret
close_swap		ENDP



;--------------------------------------------------------------------
; Write memory to swap file, free MCB.
;
; On entry:
;	BX = swap file handle
;
; Returns:
;	Carry clear on success, set on error.
;--------------------------------------------------------------------

write_mem		PROC
	push cx dx si di ds es bp

	mov bp, ds			; save program data segment
	sub di, di			; clear table index

@@next_entry:
	cmp di, [swap_count]		; compare entry count
	jae @@exit			; all done

	mov ax, SIZE MEMSAVE
	mul di
	add ax, OFFSET swap_table	; point to swap table entry
	mov si, ax

	mov dx, [si].memseg		; get save segment
	mov ax, [si].saveparas		; get paragraph count
	mov es, dx			; save MCB segment
	mov ds, dx			; set write segment

	sub dx, dx
	mov cx, $READ_WRITE_PARAS
	div cx				; get loop count
	mov si, ax			; save loop count
	or dx, dx
	jz @@block_loop			; no remainder to write

	push dx
	mov ax, $PARA_SIZE		; get remainder bytes and clear DX
	mul dx
	mov cx, ax			; set count to remainder
	mov ax, $DOS_FILE_WRITE		; DOS write to file
	int $DOS_INT
	pop dx
	jc @@exit

	cmp ax, cx			; check write count
	clc				; clear carry - not an error
	js @@exit

	mov ax, ds			; get write segment
	add ax, dx			; update write segment
	mov ds, ax
	sub dx, dx			; set write offset

@@block_loop:
	or si, si
	jz @@end_loop			; no blocks to write

	mov cx, $READ_WRITE_PARAS * $PARA_SIZE

@@next_block:
	mov ax, $DOS_FILE_WRITE		; DOS write to file
	int $DOS_INT
	jc @@exit

	cmp ax, cx			; check write count
	clc				; clear carry - not an error
	js @@exit

	mov ax, ds			; update write segment
	add ax, $READ_WRITE_PARAS
	mov ds, ax
	dec si
	jnz @@next_block

@@end_loop:
	mov es:[0].owner, 0		; mark block free
	mov ds, bp			; restore program data segment

	inc di				; increment write count
	jmp SHORT @@next_entry

@@exit:
	mov [save_count], di		; save number of swapped entries

	pop bp es ds di si dx cx
	ret
write_mem		ENDP



;--------------------------------------------------------------------
; Read memory from swap file.
;
; On entry:
;	BX = swap file handle
;
; Returns:
;	Carry clear on success, set on error.
;--------------------------------------------------------------------

read_mem		PROC
	push cx dx si di ds es

	sub di, di			; clear table index

@@next_entry:
	cmp di, [save_count]		; compare saved MCB count
	jae @@exit			; all done

	mov ax, SIZE MEMSAVE
	mul di
	add ax, OFFSET swap_table	; point to swap table entry
	mov si, ax

	mov dx, [si].memseg		; get save segment
	mov ax, [si].saveparas		; get paragraph count
	push ds
	pop es				; save program data segment in ES
	mov ds, dx			; set read segment

	sub dx, dx
	mov cx, $READ_WRITE_PARAS
	div cx				; get loop count
	mov si, ax			; save loop count
	or dx, dx
	jz @@block_loop;		; no remainder to read

	push dx
	mov ax, $PARA_SIZE		; get remainder bytes and clear DX
	mul dx
	mov cx, ax			; set count to remainder
	mov ax, $DOS_FILE_READ		; DOS read from file
	int $DOS_INT
	pop dx
	jc @@exit

	cmp ax, cx			; check read count
	jb @@exit			; exit with carry set

	mov ax, ds			; get read segment
	add ax, dx			; update read segment
	mov ds, ax
	sub dx, dx			; set read offset

@@block_loop:
	or si, si
	jz @@end_loop			; no blocks to read

	mov cx, $READ_WRITE_PARAS * $PARA_SIZE

@@next_block:
	mov ax, $DOS_FILE_READ		; DOS read from file
	int $DOS_INT
	jc @@exit

	cmp ax, cx			; check read count
	jb @@exit			; exit with carry set

	mov ax, ds			; update write segment
	add ax, $READ_WRITE_PARAS
	mov ds, ax
	dec si
	jnz @@next_block

@@end_loop:
	push es				; restore data segment
	pop ds
	
	inc di				; increment read count
	jmp SHORT @@next_entry

@@exit:
	pop es ds di si dx cx
	ret
read_mem		ENDP



;--------------------------------------------------------------------
; Spawn child process.
;
; On entry:
;	Nothing
; Returns:
;	Carry set with error code on failure, carry clear with
;	child result code on success.
;--------------------------------------------------------------------

spawn_proc		PROC
	push bx cx si di

	push ds
	pop es
	mov di, OFFSET cmd_line		; point to command line
	mov cx, [cmd_length]
	mov al, ' '			; search for space before arguments
	repne scasb

	dec di
	sub al, al
	stosb				; terminate command line

	mov si, di			; save pointer to arguments

	mov di, OFFSET cmd_tail		; copy command tail
	mov ax, cx
	inc ax				; add size of leading space
	stosb				; save tail size

	mov al, ' '			; start tail with space
	stosb
	rep movsb			; copy command tail

	mov al, 0dh			; end with carriage return
	stosb

	mov bx, OFFSET exec_param	; point to parameter block

	mov ax, [psp_seg]
	mov es, ax
	mov ax, es:[$ENV_SEG_OFFSET]	; get environment segment
	mov [bx].env_addr, ax		; set child environment
	
	mov ax, OFFSET cmd_tail		; point to command tail
	mov WORD PTR [bx].cmd_addr, ax
	mov WORD PTR [bx].cmd_addr + 2, ds

	mov ax, OFFSET blank_fcb	; point to blank FCB
	mov WORD PTR [bx].fcb_addr_1, ax
	mov WORD PTR [bx].fcb_addr_1 + 2, ds
	add ax, $FCB_SIZE
	mov WORD PTR [bx].fcb_addr_2, ax
	mov WORD PTR [bx].fcb_addr_2 + 2, ds

	push ds
	pop es
	mov dx, OFFSET cmd_line
	mov ax, $DOS_EXEC		; DOS exec function
	int $DOS_INT
	jc @@exit			; exit with error code

	mov ax, $DOS_EXEC_RETURN	; DOS get execute return code
	int $DOS_INT

@@exit:
	mov [spawn_code], ax		; save spawn result code

	pop di si cx bx
	ret
spawn_proc		ENDP



;--------------------------------------------------------------------
; Swap program memory to disk and spawn subprocess.
;
; On entry:
;	Called with near pointer to command line string.  Command
;	must contain full path of program to spawn, plus any
;	command line arguments.
;
; Returns:
;	Exit code of subprocess
;--------------------------------------------------------------------
	PUBLIC C	swap_spawn
swap_spawn	PROC C
	ARG	cmd:PTR BYTE		; pointer to command line

	push bx cx dx di si es ds	; save registers

IF @DataSize
	les di, [cmd]			; far pointer to command line
ELSE
	push ds
	pop es
	mov di, [cmd]			; near pointer to command line
ENDIF

	mov cx, $COMMAND_LENGTH		; maximum length of command line
	mov al, 0			; search for command terminator
	repne scasb

	mov bx, cx			; save remainder length
	sub cx, $COMMAND_LENGTH		; get command length
	neg cx

IF @DataSize
	lds si, [cmd]			; point to command line
ELSE
	mov si, [cmd]			; point to command line
ENDIF
	push cs
	pop es
	mov di, OFFSET cmd_line
	rep movsb			; copy command line

	mov cx, bx
	sub al, al
	rep stosb			; clear rest of command line

	push cs				; make CS data addressable
	pop ds

	sub bx, $COMMAND_LENGTH
	neg bx
	mov [cmd_length], bx		; save command line length

	cli
	mov [dos_ss], ss		; set up local stack
	mov [dos_bp], bp
	mov [dos_sp], sp
	mov ax, ds
	mov ss, ax
	mov bp, [end_stack]
	mov sp, [end_stack]
	sti

	mov ax, $DOS_GET_PSP		; DOS get PSP function
	int $DOS_INT
	mov [psp_seg], bx		; save local copy of PSP
	mov es, bx			; set ES to PSP segment

	call relocate_env		; relocate environment
	call split_psp			; shrink program memory block
	call build_table		; build swap table

	mov dx, OFFSET swap_file	; point to swap file name
	call create_swap		; create swap file
	jc swap_open_err

	call write_mem			; write memory to swap file
	jc write_mem_err

	call close_swap			; close swap file
	mov ax, 1
	call lock_swap			; lock swap file

	call save_dir			; save current directory
	call spawn_proc			; spawn child process
	call restore_dir		; restore current directory

	mov dx, OFFSET swap_file
	call open_swap			; open swap file for read
	jc swap_fatal

	call read_mem			; read memory from swap file
	jc swap_fatal

	call close_swap			; close swap file
	call delete_swap		; delete swap file

swap_open_err:
write_mem_err:
	mov bx, [psp_seg]
	call restore_psp		; rejoin program memory block
	call restore_env		; restore old environment

	cli
	mov ss, [dos_ss]		; restore old stack
	mov bp, [dos_bp]
	mov sp, [dos_sp]
	sti

	mov ax, [spawn_code]		; return exit code

@@exit:
	pop ds es si di dx cx bx
	ret

swap_fatal:
	mov dx, OFFSET file_err_msg
	mov ax, $DOS_PRINT_STRING	; DOS print string function
	int $DOS_INT

	mov ax, $DOS_EXIT		; DOS exit program function
	int $DOS_INT

swap_spawn	ENDP



;-------------------- environment and program tail --------------------

EVENDATA
new_env		DB ($MAX_ENV_SIZE + (2 * $PARA_SIZE)) DUP (0)
tail_pad	DB (SIZE MCB + $PARA_SIZE) DUP (0)	; space for tail MCB


END
